# Copyright 2016 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import time
import uuid

from ovs.db import idl as ovs_idl

from networking_ovn.common import constants as ovn_const
from networking_ovn.ovsdb import ovsdb_monitor
from networking_ovn.tests.unit import test_ovn_plugin

OVN_PROFILE = ovn_const.OVN_PORT_BINDING_PROFILE


OVN_NB_SCHEMA = {
    "name": "OVN_Northbound", "version": "2.0.1",
    "tables": {
        "Logical_Port": {
            "columns": {
                "name": {"type": "string"},
                "type": {"type": "string"},
                "addresses": {"type": {"key": "string",
                                       "min": 0,
                                       "max": "unlimited"}},
                "port_security": {"type": {"key": "string",
                                           "min": 0,
                                           "max": "unlimited"}},
                "up": {"type": {"key": "boolean", "min": 0, "max": 1}}},
            "indexes": [["name"]],
            "isRoot": False,
        },
        "Logical_Switch": {
            "columns": {"name": {"type": "string"}},
            "indexes": [["name"]],
            "isRoot": True,
        }
    }
}


class TestOvnIdlNotifyHandler(test_ovn_plugin.OVNPluginTestCase):

    def setUp(self):
        super(TestOvnIdlNotifyHandler, self).setUp()
        helper = ovs_idl.SchemaHelper(schema_json=OVN_NB_SCHEMA)
        helper.register_all()
        self.idl = ovsdb_monitor.OvnIdl(self.plugin, "remote", helper)
        self.idl.lock_name = self.idl.event_lock_name
        self.idl.has_lock = True
        self.lp_table = self.idl.tables.get('Logical_Port')
        self.plugin.set_port_status_up = mock.Mock()
        self.plugin.set_port_status_down = mock.Mock()

    def _test_lport_helper(self, event, new_row_json, old_row_json=None,
                           table=None):
        row_uuid = str(uuid.uuid4())
        if not table:
            table = self.lp_table
        lp_row = ovs_idl.Row.from_json(self.idl, table,
                                       row_uuid, new_row_json)
        if old_row_json:
            old_row = ovs_idl.Row.from_json(self.idl, table,
                                            row_uuid, old_row_json)
        else:
            old_row = None
        self.idl.notify(event, lp_row, updates=old_row)
        # sleep for a second so that the notify handler green thread
        # handles the notify event
        time.sleep(1)

    def test_lport_up_create_event(self):
        row_data = {"up": True, "name": "foo-name"}
        self._test_lport_helper('create', row_data)
        self.plugin.set_port_status_up.assert_called_once_with("foo-name")
        self.assertFalse(self.plugin.set_port_status_down.called)

    def test_lport_down_create_event(self):
        row_data = {"up": False, "name": "foo-name"}
        self._test_lport_helper('create', row_data)
        self.plugin.set_port_status_down.assert_called_once_with("foo-name")
        self.assertFalse(self.plugin.set_port_status_up.called)

    def test_lport_up_not_set_event(self):
        row_data = {"up": ['set', []], "name": "foo-name"}
        self._test_lport_helper('create', row_data)
        self.assertFalse(self.plugin.set_port_status_up.called)
        self.assertFalse(self.plugin.set_port_status_down.called)

    def test_unwatch_logical_port_create_events(self):
        self.idl.unwatch_logical_port_create_events()
        row_data = {"up": True, "name": "foo-name"}
        self._test_lport_helper('create', row_data)
        self.assertFalse(self.plugin.set_port_status_up.called)
        self.assertFalse(self.plugin.set_port_status_down.called)

        row_data["up"] = False
        self._test_lport_helper('create', row_data)
        self.assertFalse(self.plugin.set_port_status_up.called)
        self.assertFalse(self.plugin.set_port_status_down.called)

    def test_lport_up_update_event(self):
        new_row_json = {"up": True, "name": "foo-name"}
        old_row_json = {"up": False}
        self._test_lport_helper('update', new_row_json,
                                old_row_json=old_row_json)
        self.plugin.set_port_status_up.assert_called_once_with("foo-name")
        self.assertFalse(self.plugin.set_port_status_down.called)

    def test_lport_down_update_event(self):
        new_row_json = {"up": False, "name": "foo-name"}
        old_row_json = {"up": True}
        self._test_lport_helper('update', new_row_json,
                                old_row_json=old_row_json)
        self.plugin.set_port_status_down.assert_called_once_with("foo-name")
        self.assertFalse(self.plugin.set_port_status_up.called)

    def test_lport_up_update_event_no_old_data(self):
        new_row_json = {"up": True, "name": "foo-name"}
        self._test_lport_helper('update', new_row_json,
                                old_row_json=None)
        self.assertFalse(self.plugin.set_port_status_up.called)
        self.assertFalse(self.plugin.set_port_status_down.called)

    def test_lport_down_update_event_no_old_data(self):
        new_row_json = {"up": False, "name": "foo-name"}
        self._test_lport_helper('update', new_row_json,
                                old_row_json=None)
        self.assertFalse(self.plugin.set_port_status_up.called)
        self.assertFalse(self.plugin.set_port_status_down.called)

    def test_lport_other_column_update_event(self):
        new_row_json = {"up": False, "name": "foo-name",
                        "addresses": ["10.0.0.2"]}
        old_row_json = {"addresses": ["10.0.0.3"]}
        self._test_lport_helper('update', new_row_json,
                                old_row_json=old_row_json)
        self.assertFalse(self.plugin.set_port_status_up.called)
        self.assertFalse(self.plugin.set_port_status_down.called)

    def test_notify_other_table(self):
        new_row_json = {"name": "foo-name"}
        self._test_lport_helper('create', new_row_json,
                                table=self.idl.tables.get("Logical_Switch"))
        self.assertFalse(self.plugin.set_port_status_up.called)
        self.assertFalse(self.plugin.set_port_status_down.called)

    def test_notify_no_ovsdb_lock(self):
        self.idl.has_lock = False
        self.idl.is_lock_contended = True
        self.idl.notify_handler.notify = mock.Mock()
        self.idl.notify("create", mock.ANY)
        self.assertFalse(self.idl.notify_handler.notify.called)

    def test_notify_ovsdb_lock_not_yet_contended(self):
        self.idl.has_lock = False
        self.idl.is_lock_contended = False
        self.idl.notify_handler.notify = mock.Mock()
        self.idl.notify("create", mock.ANY)
        self.assertTrue(self.idl.notify_handler.notify.called)
